local B, L, T = unpack(select(2, ...))
local Search = LibStub:GetLibrary('LibItemSearch-1.0')

--[[//////////////////////////////////////////////////

    HELPER FUNCTIONS

    Anything that needs repetitive actions or calculations
    Tips: 
        -- Multiplication are faster than division
        -- Local is faster than global. Re-assign global as a new local if used a lot

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    -- Table pool for recycling tables
    local tables = setmetatable({}, {__mode = "kv"}) -- Weak table

    -- Get a new table
    function B.newtable()
        local t = next(tables) or {}

        tables[t] = nil
        return t
    end

    function B.newsearchtable()
        local t = B.newtable()

        setmetatable(t, {
        __call = function(t, loc, num, name, returnTable)
            local x

            for x = 1, #t do
                if t[x].loc == loc and t[x].num == num and t[x].name == name then
                    return t[x]
                end
            end

            if returnTable == false then return end

            t[#t + 1] = B.newtable()
            return t[#t]
        end
        })
        return t
    end

    -- Delete table and add to table pool
    function B.deltable(t)

        if type(t) == "table" then
            for k, v in pairs(t) do
                if type(v) == "table" then B.deltable(v) end
                t[k] = nil
            end
            
            setmetatable(t, nil)
            t[true] = true
            t[true] = nil
            tables[t] = true
        end
        return nil
    end

    -- tcopy: recursively copy contents of one table to another.  from wowwiki
    function B.copytable(to, from)
        -- "to" must be a table (possibly empty)
        for k, v in pairs(from) do

            if type(v) == "table" then
                to[k] = {}
                B.copytable(to[k], v)
            else
                to[k] = v
            end
        end
    end

    function B.strstart(string, start)
       return string.sub(string, 1, string.len(start)) == start
    end

    function B.template(server, player, faction)

        T.ACTIVE_TABLE = {['server'] = server, ['player'] = player, ['faction'] = faction}
        
        if not BWsavedtable then
            BWsavedtable = {}
            BWsavedtable['iconsizeslider'] = 32
            BWsavedtable['iconsize'] = 32
            BWsavedtable['iconrowslider'] = 12
            BWsavedtable['iconrow'] = 12
            BWsavedtable['minimize'] = false
            BWsavedtable['bankbagstoggle'] = true
            BWsavedtable['bankbagtooltiptoggle'] = true
        end

        if not BWsavedtable[server] then BWsavedtable[server] = {} end

        BWsavedtable[server][player.."."..string.sub(faction, 1, 1)] = {['bank'] = {{['slots'] = NUM_BANKGENERIC_SLOTS}}, ['reagent'] = {{}},
                                                                        ['void'] = {{}, {}}, ['guild'] = {}, ['player'] = {[0] = {['slots'] = NUM_BAG_SLOTS}, [-1] = {}},
                                                                        ['auctions'] = {{}}}
        B.load(server, player, faction)
    end

    function B.load(name, player, faction)
        local text

        T.ACTIVE_TABLE = {['server'] = name, ['player'] = player, ['faction'] = faction}
        -- Validate settings added at a later time
        if not BWsavedtable[name][player.."."..string.sub(faction, 1, 1)]['player'] then BWsavedtable[name][player.."."..string.sub(faction, 1, 1)]['player'] = {} end
        if not BWsavedtable[name][player.."."..string.sub(faction, 1, 1)]['auctions'] then BWsavedtable[name][player.."."..string.sub(faction, 1, 1)]['auctions'] = {{}} end
        if BWsavedtable['bankbagstoggle'] == nil then BWsavedtable['bankbagstoggle'] = true end
        if BWsavedtable['bankbagtooltiptoggle'] == nil then BWsavedtable['bankbagtooltiptoggle'] = true end

        B.deltable(T.BW.items)
        T.BW.items = B.newsearchtable()

        if B.IsCurrentPlayer() then
            T.F.mainframe.character.artifact:Enable()
        else
            T.F.mainframe.character.artifact:Disable()
        end

        B.BankUpdate()
        B.ReagentUpdate()
        B.VoidUpdate()
        B.GuildUpdate()
        B.PlayerUpdate()
        B.PlayerEquipmentUpdate()

        text = T.F.mainframe.searchbar:GetText()
        if text and text ~= "" then
            T.F.mainframe.searchbar:SetText("")
            T.F.mainframe.searchbar:SetText(text)
        else
            B.Search(false)
        end
    end

    function B.IsCurrentPlayer()

        local player = UnitName("player")
        local server = GetRealmName()
        local faction = UnitFactionGroup("player")

        if T.ACTIVE_TABLE.server == server and T.ACTIVE_TABLE.player == player and T.ACTIVE_TABLE.faction then
            return true
        end
        return false
    end

    function B.GetPlayerBank(bank, current)

        if not bank then return end
        if current then
            local player = UnitName("player")
            local server = GetRealmName()
            local faction = UnitFactionGroup("player")
            return BWsavedtable[server][player.."."..string.sub(faction, 1, 1)][bank]
        end

        return BWsavedtable[T.ACTIVE_TABLE.server][T.ACTIVE_TABLE.player.."."..string.sub(T.ACTIVE_TABLE.faction, 1, 1)][bank]
    end


    local stack = -1
    local messagestr = ""

    function B.print(message, level)
        local filler

        if BWdebug or (level == 1) then

            if stack == -1 then messagestr = T.MSGC.DEFAULT.."Bank's Watch|r: " end

            if not message then
                DEFAULT_CHAT_FRAME:AddMessage(messagestr..tostring(message))
                return false 
            end

            stack = stack + 1

            if type(message) == "table" then
                
                filler = string.rep(". . ", stack)
                DEFAULT_CHAT_FRAME:AddMessage(messagestr.."{table} -->")
                for k, v in pairs(message) do
                    messagestr = filler.."["..k.."] = "
                    B.print(v)
                end
            
            elseif type(message) == "userdata" then

            elseif type(message) == "function" then
                DEFAULT_CHAT_FRAME:AddMessage(messagestr.." A function(self)")
            elseif type(message) == "boolean" then
                DEFAULT_CHAT_FRAME:AddMessage(messagestr..tostring(message))
            else
                DEFAULT_CHAT_FRAME:AddMessage(messagestr..tostring(message))
            end
            
            stack = stack - 1
       end
    end

    function B.find(text, itemOrButton)

        if strfind(strlower(ARTIFACT_POWER), text) and IsArtifactPowerItem(B.GetItemIdFromLink(itemOrButton.link)) then
            return true
        elseif strfind(strlower(TOOLTIP_BATTLE_PET), text) and strfind(itemOrButton.link, "|Hbattlepet:") then
            return true
        else
            return Search:Find(itemOrButton, text)
        end
    end

    function B.GetCursorScaledPosition()
        local scale, x, y = UIParent:GetScale(), GetCursorPosition()
        return x / scale, y / scale
    end

    function B.SetFrameLocation()
        local x, y = T.F.mainframe.optionbutton:GetRect()
        -- Only hide when cursor exits at the top or bottom of the option menu
        if (x < (GetScreenWidth()/2)) and (y > (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 1 then B.I.AnchorBagsTopLeft() end
        elseif (x < (GetScreenWidth()/2)) and (y < (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 2 then B.I.AnchorBagsBottomLeft() end
        elseif (x > (GetScreenWidth()/2)) and (y > (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 3 then B.I.AnchorBagsTopRight() end
        elseif (x > (GetScreenWidth()/2)) and (y < (GetScreenHeight()/2)) then
            if T.F.mainframe.layout ~= 4 then B.I.AnchorBagsBottomRight() end
        end
        B.ScrollbarUpdate()
    end

    function B.RowsToFit()

        local ourheight = math.ceil(T.F.mainframe.search.list:GetHeight()) - T.FRAMEWHITESPACE
        return math.floor(ourheight / T.BUTTON_HEIGHT)
    end

    function B.HideAll()

        T.F.mainframe.bank:Hide()
        T.F.mainframe.reagent:Hide()
        T.F.mainframe.void:Hide()
        T.F.mainframe.guild:Hide()
        T.F.mainframe.search:Hide()
        T.F.mainframe.character:Hide()
    end

    local ilvl_pattern = ITEM_LEVEL:gsub("%%d", "(%%d+)")
    local tooltip = CreateFrame("GameTooltip", "BW_tooltip", nil, "GameTooltipTemplate")

    function B.DetectILVL(itemlink)
        local text, ilvl

        tooltip:SetOwner(UIParent, "ANCHOR_NONE")
        tooltip:SetHyperlink(itemlink)

        text = BW_tooltipTextLeft2:GetText()
        if text then ilvl = tonumber(text:match(ilvl_pattern)) end

        tooltip:ClearLines()
        return ilvl or nil
    end

    function B.DetectAP(bag, slot)
        local text, ap, k

        tooltip:SetOwner(UIParent, "ANCHOR_NONE")
        tooltip:SetBagItem(bag, slot)

        for k = tooltip:NumLines(), 2, -1 do
            ap = (_G["BW_tooltipTextLeft"..k]:GetText() or ""):match("(%d*%.?%d+ %a+ "..ARTIFACT_POWER..")")
            if ap then
                ap = ap:gsub(",", "")
                tooltip:ClearLines()
                return ap
            end
        end

        tooltip:ClearLines()
        return 0
    end

    function B.GetGuildPetLink(page, slot)
        local speciesID, level, breedQuality, maxHealth, power, speed, name = tooltip:SetGuildBankItem(page, slot)
        local link = string.format("%s|Hbattlepet:%s:%s:%s:%s:%s:%s:|h[%s]|h|r", ITEM_QUALITY_COLORS[breedQuality].hex, speciesID, level, breedQuality, maxHealth, power, speed, name)
        tooltip:ClearLines()

        return link
    end

    function B.GetPetInfo(itemlink)
        local speciesID, level, breedQuality, maxHealth, power, speed, name, icon

        _, speciesID, level, breedQuality, maxHealth, power, speed, name = strsplit(":", itemlink)
        name, icon = C_PetJournal.GetPetInfoBySpeciesID(speciesID)
        return name, itemlink, tonumber(breedQuality), nil, nil, nil, nil, nil, nil, icon -- Keep same structure as getiteminfo
    end

    function B.GetKeystoneInfo(itemlink)
        local _, _, _, _, _, _, _, _, _, _, _, _, _, _, name = string.find(itemlink, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")
        return name, itemlink, 4, nil, nil, nil, nil, nil, nil, 525134 -- Keep same structure as getiteminfo
    end

    function B.GetItemIdFromLink(itemlink)
        local itemstring = strmatch(itemlink, "item[%-?%d:]+")
        if not itemstring then return nil end
        local _, itemid = strsplit(":", itemstring)

        return itemid
    end

    function B.GetQualityFromLink(itemlink)
        local _, _, color = strfind(itemlink, "|?c?f?f?(%x*)|")
        return T.QUALITY[color]
    end

    function B.GetBag(button)
        local total, lasttotal, page
        local id = button:GetID()
        local player = T['player']

        total = 0
        lasttotal = 0

        for page = 0, #player do
            lasttotal = total
            total = total + (player[page] and player[page]['slots'] or 0)
            if total >= id then return page, page == 0 and id or id - lasttotal end
        end
    end

    function B.GetBankBag(button)
        local total, lasttotal, page
        local id = button:GetID()
        local bank = T['bank']

        total = 0
        lasttotal = 0

        for page = 1, #bank do
            lasttotal = total
            total = total + (bank[page] and bank[page]['slots'] or 0)
            if total >= id then return page == 1 and BANK_CONTAINER or page, page == 1 and id or id - lasttotal end
        end
    end

    function B.GetSlotID(bag, slot)
        local total, lasttotal, page
        local bank = T['bank']

        total = 0

        for page = 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do
            if bank[page] then
                
                if page == bag then return total + slot end
                total = total + bank[page]['slots']
            end
        end
    end

    function B.IsGuildBank(button)
        local name = button:GetName()
        if name and strfind(name, "GuildBankColumn%d+Button%d+") then
            return true
        end
        return false
    end

    function B.IsGenericBank(button)
        local name = button:GetName()
        if name and strfind(name, "BankFrameItem%d+") then
            return true
        end
        return false
    end

    function B.IsBankBag(button)
        local name = button:GetName()
        if name and strfind(name, "ContainerFrame%d+Item%d+") then
            local bagid = button:GetParent():GetID()
            if bagid >= (NUM_BAG_SLOTS + 1) and bagid <= (NUM_BAG_SLOTS + NUM_BANKBAGSLOTS) then
                return true
            end
        end
        return false
    end

    function B.IsReagentBank(button)
        local name = button:GetName()
        if name and strfind(name, "ReagentBankFrameItem%d+") then
            return true
        end
        return false
    end

    function B.IsVoidBank(button)
        local name = button:GetName()
        if name and strfind(name, "VoidStorageStorageButton%d+") then
            local id = strmatch(name, ("%d+"))
            button:SetID(tonumber(id)) -- Blizzard doesn't set id, so I will
            return true
        end
        return false
    end

    function B.IsPlayerBag(button)
        local name = button:GetName()
        if name and strfind(name, "ContainerFrame%d+Item%d+") then
            local bagid = button:GetParent():GetID()
            if bagid > -1 and bagid <= NUM_BAG_SLOTS then
                return true
            end
        end
        return false
    end

    function B.ToggleAllBags()
        local i

        --FRAME_THAT_OPENED_BAGS = nil;
        CloseBackpack();
        
        for i = 1, NUM_BAG_FRAMES, 1 do
            CloseBag(i)
        end
    end
